using Toybox.Application.Properties;
using Toybox.Graphics as Gfx;
using Toybox.Math;
using Toybox.System as Sys;
using Toybox.Time;
using Toybox.Time.Gregorian;

import Toybox.Application.Storage;
import Toybox.Lang;
import Toybox.StringUtil;
import Toybox.WatchUi;

module C
{

//
//  Define my own colors since Garmin is a little color
//  blind (0xff00ff is magenta, not pink).
//
public const BLACK   = 0x000000;
public const WHITE   = 0xffffff;
public const GRAY    = 0x888888;
public const RED     = 0xff0000;
public const GREEN   = 0x00ff00;
public const BLUE    = 0x0000ff;
public const CYAN    = 0x00ffff;
public const MAGENTA = 0xff00ff;
public const YELLOW  = 0xffff00;

//
//  Main view statews
//
public const MAIN_SCAN          = 0;
public const MAIN_WAIT          = 1;
public const MAIN_ENUMERATE     = 2;
public const MAIN_SHOW          = 3;
public const MAIN_ERROR         = 4;
public const MAIN_LOG           = 5;

//
//  Intelligent outlet types
//
public const OUTLET_TPLINK = 0;
public const OUTLET_WEMO   = 1;     // must be equal to OUTLET_TPLINK + 1
public const OUTLET_ATHOME = 2;     // must be equal to OUTLET_WEMO + 1
public const OUTLET_NONE   = 3;     // must be equal to OUTLET_ATHOME + 1
public const OUTLET_ERROR  = 4;

//
//  Outlet states
//
public const STATE_OFF      = 0;
public const STATE_ON       = 1;
public const STATE_OFL      = 2;
public const STATE_UNK      = 3;

//
//  Outlet icons
//
public const ICON_PLUG   = 0;
public const ICON_STRIP  = 1;
public const ICON_SWITCH = 2;

//
//  Outlet controls
//
public const CTRL_GET       = 0;
public const CTRL_SET       = 1;
public const CTRL_ENUMERATE = 2;

//
//  Menu constants
//
public const MENU_NONE      = 0;
public const MENU_ONOFF     = 1;
public const MENU_DEVICE    = 100;

//
//  Watch face dimensions
//
public var cx;
public var cy;
public var radius;

//
//  Common data
//
public var state = MAIN_SCAN;
public var alert_msg = "Wait internet";
public var outlets as Array<Dictionary> = new [0];
public var dev_list;
public var enumerating = OUTLET_TPLINK;
public var onoff_state;
public var onoff_ood;

public var tplink;
public var wemo;
public var athome;

//
//  icons
//

public var icons;
public const icons_sm = [
    Rez.Drawables.outlet_off_sm, Rez.Drawables.outlet_on_sm, Rez.Drawables.outlet_ofl_sm,
    Rez.Drawables.strip_off_sm,  Rez.Drawables.strip_on_sm,  Rez.Drawables.strip_ofl_sm,
    Rez.Drawables.switch_off_sm, Rez.Drawables.switch_on_sm, Rez.Drawables.switch_ofl_sm,
];
public const icons_med = [
    Rez.Drawables.outlet_off_med, Rez.Drawables.outlet_on_med, Rez.Drawables.outlet_ofl_med,
    Rez.Drawables.strip_off_med,  Rez.Drawables.strip_on_med,  Rez.Drawables.strip_ofl_med,
    Rez.Drawables.switch_off_med, Rez.Drawables.switch_on_med, Rez.Drawables.switch_ofl_med,
];
public const icons_lg = [
    Rez.Drawables.outlet_off_lg, Rez.Drawables.outlet_on_lg, Rez.Drawables.outlet_ofl_lg,
    Rez.Drawables.strip_off_lg,  Rez.Drawables.strip_on_lg,  Rez.Drawables.strip_ofl_lg,
    Rez.Drawables.switch_off_lg, Rez.Drawables.switch_on_lg, Rez.Drawables.switch_ofl_lg,
];

//
//  Settings
//
public var settings as Dictionary<Object, String or Number or Boolean or Array<String>> = {
    "offline"      =>  true,
    "sort_dir"     =>  1,
    "tp_svr"       =>  "",
    "tp_usr"       =>  "",
    "tp_pwd"       =>  "",
//    "wemo_svr"     =>  "",
//    "wemo_usr"     =>  "",
//    "wemo_pwd"     =>  "",
    "athome_svr"   =>  "",
    "athome_usr"   =>  "",
    "athome_pwd"   =>  "",
};

function get_settings()
{

    var ver = P.get_prop("version", "1.0");

    settings["offline"] = P.get_prop("offline", true);

    switch (P.get_prop("sort_dir", 1)) {

    case 0:     // descending
        settings["sort_dir"] = -1;
        break;

    case 1:     // ascending
        settings["sort_dir"] = 1;
        break;

    default:    // no sort at all
        settings["sort_dir"] = 0;
        break;

    }

    settings["tp_svr"] = P.get_prop("tp_svr", "");
    settings["tp_usr"] = P.get_prop("tp_usr", "");
    settings["tp_pwd"] = P.get_prop("tp_pwd", "");

//    settings["wemo_svr"] = P.get_prop("wemo_svr", "");
//    settings["wemo_usr"] = P.get_prop("wemo_usr", "");
//    settings["wemo_pwd"] = P.get_prop("wemo_pwd", "");

    settings["athome_svr"] = P.get_prop("athome_svr", "");
    settings["athome_usr"] = P.get_prop("athome_usr", "");
    settings["athome_pwd"] = P.get_prop("athome_pwd", "");

    if (L.debug != L.DBG_NONE) {
        settings["offline"] = true;
        settings["tp_svr"] = "https://wap.tplinkcloud.com";
        settings["tp_usr"] = "";
        settings["tp_pwd"] = "";
        settings["athome_svr"] = "n0ano.com/cgi-bin/garmin.pl";
        settings["athome_usr"] = "";
        settings["athome_pwd"] = "";
    }

    L.pr_json(L.DBG_BASIC, "settings " + ver + ":", settings);
}

//
//  Garmin doesn't provide an API to generate a UUID
//  and the UUID provided to a TpLink API doesn't have
//  be that unique.  To simplify things we'll use a
//  template from a working UUID and just provide make
//  the first 32 bits randome, should be sufficient.
function fake_uuid()
{
    var r;

    Math.srand(Time.now().value());
    r = Math.rand();
    r = r.format("%8.8x") + "-a5b5-4774-86e6-15338b4f0149";
    return r;
}

function set_fill(dc, color)
{

    if (dc has :setFill) {
        dc.setFill(color);
    } else {
        dc.setColor(color, Gfx.COLOR_TRANSPARENT);
    }
}

function set_alert(msg)
{

    alert_msg = msg;
    WatchUi.requestUpdate();
}

function get_ood(size)
{

    if (size <= 0) {
        return Rez.Drawables.ood_sm; }
    else if (size < 150) {
        return Rez.Drawables.ood_med; }
    return Rez.Drawables.ood_lg;
}

function get_icon(icon, onoff, icons as Array<Number>)
{

    //
    //  This code assumes there are 3 on/off states (0, 1, 2)
    //  and the icons come outlet first, then strip, and
    //  finally switch
    //
    switch (icon) {

    case ICON_SWITCH:
        onoff += 3;
        // fall thru
    case ICON_STRIP:
        onoff += 3;
        // fall thru
    case ICON_PLUG:
        break;

    }
    return icons[onoff];
}
//
//  Note that the callback for these APIs includes 2 paramters,
//  the first is the return value and the second is either a
//  null (the return value is valid) or an error string (the
//  return value is arbitrary).
function dev_ctrl(op, dev as Dictionary, state, cb)
{

    var type = dev["type"];
    switch (type) {

    case OUTLET_TPLINK:
        tplink.ctrl(op, dev, state, cb);
        return;

    case OUTLET_WEMO:
        break;

    case OUTLET_ATHOME:
        athome.ctrl(op, dev, state, cb);
        return;

    }
    L.dbg_log(L.DBG_BASIC, "dev_ctrl:" + type + " => unknonw device type");
    cb.invoke(false);
}

function sh_progress(msg, delegate)
{

    var pb = new WatchUi.ProgressBar(msg, null);
    WatchUi.pushView(pb, delegate, WatchUi.SLIDE_IMMEDIATE);
}

function load_devs()
{
    var dev;

    outlets = Storage.getValue("Devices");
    if (outlets != null) {
        for (var i = 0; i < C.outlets.size(); i += 1) {
            dev = C.outlets[i];
            dev["ood"] = true;
        }
        return true;
    }
    return false;
}

function save_devs()
{

    Storage.setValue("Devices", outlets);
}

function state2str(s)
{

    switch (s) {

    case STATE_OFL:
        return "offline";

    case STATE_OFF:
        return "off";

    case STATE_ON:
        return "on";

    }

    return "??";
}

function outlet_sort(dir)
{
    var dev;

    for (var i = 0; i < outlets.size() - 1; i += 1) {
        for (var j = i + 1; j < outlets.size(); j += 1) {
            if (dev_comp(outlets[i], outlets[j], dir) < 0) {
                dev = outlets[i];
                outlets[i] = outlets[j];
                outlets[j] = dev;
            }
        }
    }
}

function dev_comp(a as Dictionary, b as Dictionary, dir)
{

    var n1 = a["name"];
    var n2 = b["name"];
    return compare_to(n1, n2) * dir;
}

function compare_to(s1, s2)
{
    var n;

    var b1 = s1.toUtf8Array() as Array<Number>;
    var l1 = b1.size();
    var b2 = s2.toUtf8Array() as Array<Number>;
    var l2 = b2.size();
    var len = (l1 < l2) ? l1 : l2;
    for (var i = 0; i < len; i += 1) {
        n = b2[i] - b1[i];
        if (n != 0) {
            return  n; }
    }
    if (l1 == l2) {
        return 0;
    } else if (l1 < l2) {
        return 1;
    } else {
        return -1; }
}

}
